@@ -2,13 +2,13 @@
 # copyright notices and license terms.
 import copy
 
+from collections import defaultdict
 from decimal import Decimal
-from time import *
 
 from trytond.exceptions import UserError
 from trytond.i18n import gettext
 from trytond.model import ModelSQL, ModelView, Workflow, fields
-from trytond.pool import Pool
+from trytond.pool import Pool, PoolMeta
 from trytond.pyson import (
     And, Bool, Equal, Eval, Get, If, In, Not, Or, PYSONEncoder)
 from trytond.rpc import RPC
@@ -16,18 +16,7 @@ from trytond.transaction import Transaction
 from trytond.wizard import (
     Button, StateAction, StateTransition, StateView, Wizard)
 
-_STATES = {
-    'readonly': Equal(Eval('state'), 'closed'),
-    }
-_DEPENDS = ['state']
 
-_LINE_STATES = {
-    'readonly': Or(
-        Not(Bool(Eval('journal'))),
-        (Equal(Eval('state'), 'posted'))
-        )
-    }
-_LINE_DEPENDS = ['journal', 'state']
 _ZERO = Decimal('0.0')
 
 
@@ -35,14 +24,17 @@ class Batch(Workflow, ModelSQL, ModelView):
     'Account Batch'
     __name__ = 'account.batch'
 
-    name = fields.Char('Name', required=True, states=_STATES,
+    _states = {'readonly': Eval('state') == 'closed'}
+    _depends = ['state']
+
+    name = fields.Char('Name', required=True, states=_states,
         depends=['state'])
     company = fields.Many2One('company.company', 'Company', required=True,
-        select=True, states=_STATES, domain=[
+        select=True, states=_states, domain=[
             ('id', If(Eval('context', {}).contains('company'), '=', '!='),
                 Eval('context', {}).get('company', -1)),
             ],
-        depends=_DEPENDS)
+        depends=_depends)
     journal = fields.Many2One('account.batch.journal', 'Journal',
         required=True, states={
             'readonly': Or(
@@ -51,7 +43,7 @@ class Batch(Workflow, ModelSQL, ModelView):
                 ),
         }, domain=[
             ('company', '=', Eval('context', {}).get('company', -1)),
-        ], select=True, depends=_DEPENDS + ['journal_type', 'lines'])
+        ], select=True, depends=_depends + ['journal_type', 'lines'])
     lines = fields.One2Many('account.batch.line', 'batch',
             'Batch Lines', add_remove=[],
         # XXX 'bool' object is not iterable (<type 'exceptions.TypeError'>)
@@ -64,7 +56,7 @@ class Batch(Workflow, ModelSQL, ModelView):
         #    )],
             context={
                 'journal': Eval('journal'),
-                }, states=_STATES, depends=_DEPENDS)
+                }, states=_states, depends=_depends)
     move_lines = fields.Function(fields.One2Many('account.move.line',
             None, 'Move Lines'), 'get_move_lines')
     currency_digits = fields.Function(fields.Integer('Currency Digits'),
@@ -76,6 +68,8 @@ class Batch(Workflow, ModelSQL, ModelView):
             ('closed', 'Closed'),
             ], 'State', readonly=True)
 
+    del _states, _depends
+
     @classmethod
     def __setup__(cls):
         super(Batch, cls).__setup__()
@@ -87,7 +81,7 @@ class Batch(Workflow, ModelSQL, ModelView):
                     'invisible': Eval('state') != 'open',
                     },
                 })
-        cls.__rpc__.update({'selection_journal_types': RPC(False)})
+        cls.__rpc__.update({'selection_journal_types': RPC()})
 
     def get_rec_name(self, name):
         name = ' - ' + self.name if self.name else ''
@@ -212,12 +206,20 @@ class BatchLine(ModelSQL, ModelView):
     'Account Batch Line'
     __name__ = 'account.batch.line'
 
+    _states = {
+        'readonly': Or(
+            ~Eval('journal'),
+            Eval('state') == 'posted',
+            ),
+        }
+    _depends = ['journal', 'state']
+
     sequence = fields.Char('Seq.', readonly=True,
             help='Chronological sequence of batch lines by creation date')
     code = fields.Char('Code', readonly=True,
             help='Chronological sequence of batch lines by posting date')
-    posting_text = fields.Char('Posting Text', states=_LINE_STATES,
-            depends=_LINE_DEPENDS)
+    posting_text = fields.Char('Posting Text', states=_states,
+            depends=_depends)
     batch = fields.Many2One('account.batch', 'Batch', ondelete='CASCADE',
             domain=[('journal', '=', Eval('journal'))], states={
                 'readonly': Or(
@@ -235,13 +237,16 @@ class BatchLine(ModelSQL, ModelView):
                     )
             }, depends=['state', 'journal'], select=True)
     fiscalyear = fields.Many2One('account.fiscalyear', 'Fiscal Year',
-            states=_LINE_STATES, depends=_LINE_DEPENDS)
-    date = fields.Date('Date', required=True, states=_LINE_STATES,
-        depends=_LINE_DEPENDS, select=True)
+            states=_states, depends=_depends)
+    date = fields.Date('Date', required=True, states={
+            'readonly': Or(Bool(Eval('move')),
+                Bool(_states['readonly'])),
+            },
+        depends=_depends + ['move'], select=True)
     amount = fields.Numeric('Amount', required=True,
             digits=(16, Eval('currency_digits', 2)),
-            depends=_LINE_DEPENDS+['currency_digits'],
-            states=_LINE_STATES)
+            depends=_depends + ['currency_digits'],
+            states=_states)
     account = fields.Many2One('account.account', 'Account', required=True,
         domain=[
             ('type', '!=', None),
@@ -263,8 +268,8 @@ class BatchLine(ModelSQL, ModelView):
             ('type', '!=', None),
             ('closed', '!=', True),
             ],
-        states=_LINE_STATES,
-        depends=_LINE_DEPENDS)
+        states=_states,
+        depends=_depends)
     side_account = fields.Function(fields.Selection([
                 ('', ''),
                 ('debit', 'D'),
@@ -281,20 +286,20 @@ class BatchLine(ModelSQL, ModelView):
             help="'D': Contra Account is debit side\n"
             "'C': Contra Account is credit side"),
             'get_function_fields')
-    party = fields.Many2One('party.party', 'Party', states=_LINE_STATES,
-        depends=_LINE_DEPENDS)
+    party = fields.Many2One('party.party', 'Party', states=_states,
+        depends=_depends)
     is_cancelation_move = fields.Boolean('Cancelation',
-        states=_LINE_STATES, depends=_LINE_DEPENDS,
+        states=_states, depends=_depends,
         help='Create an inverted move for an inverted amount '
         'to compensate for a same move with original amount.')
     move = fields.Many2One('account.move', 'Account Move', readonly=True,
             select=True)
     reference = fields.Char('Reference',
-            states=_LINE_STATES, depends=_LINE_DEPENDS,
+            states=_states, depends=_depends,
             help='This field collects references to external documents, '
             'like voucher or receipt numbers.')
-    maturity_date = fields.Date('Maturity Date', states=_LINE_STATES,
-            depends=_LINE_DEPENDS, help='Date to pay the amount of the '
+    maturity_date = fields.Date('Maturity Date', states=_states,
+            depends=_depends, help='Date to pay the amount of the '
             'batch line at least.')
     currency_digits = fields.Function(fields.Integer('Currency Digits'),
             'get_function_fields')
@@ -324,12 +329,12 @@ class BatchLine(ModelSQL, ModelView):
                         ['revenue', 'expense']),
                     And(
                         Not(In(Get(Eval('_parent_batch', {}), 'journal_type'),
-                                ['revenue', 'expense'])),
-                        Bool(_LINE_STATES['readonly']))
+                            ['revenue', 'expense'])),
+                        Bool(_states['readonly']))
                     )
                 ),
             },
-        depends=_LINE_DEPENDS + ['party', 'tax'])
+        depends=_depends + ['party', 'tax'])
     tax = fields.Many2One('account.tax', 'Tax',
         domain=[
             ('parent', '=', None),
@@ -344,14 +349,17 @@ class BatchLine(ModelSQL, ModelView):
             ],
         states={'readonly':
                 Or(Bool(Eval('invoice')),
-                    _LINE_STATES['readonly'],
+                    _states['readonly'],
                 )
             },
-        depends=_LINE_DEPENDS + ['invoice', 'date'])
+        depends=_depends + ['invoice', 'date'])
+
+    del _states, _depends
 
     @classmethod
     def __setup__(cls):
         super(BatchLine, cls).__setup__()
+        cls.__rpc__.update({'post': RPC(readonly=False, instantiate=0)})
         # batch is in excludes to avoid general update of all moves.
         cls._check_modify_exclude = {'code', 'batch', 'sequence'}
         cls._order[0] = ('id', 'DESC')
@@ -592,8 +600,9 @@ class BatchLine(ModelSQL, ModelView):
         Date = pool.get('ir.date')
         today = Date.today()
 
+        context = Transaction().context
         date = self.date or today
-        company_id = Transaction().context.get('company', False)
+        company_id = context.get('company', False)
         if not company_id:
             return
         fiscalyear_for_date = Fiscalyear.find(company_id, date=date,
@@ -606,14 +615,14 @@ class BatchLine(ModelSQL, ModelView):
             self.account = self.batch.journal.account
             self.contra_account = None
         # check if date is valid for the period/fiscalyear
-        if Transaction().context.get('period'):
-            period = Period(Transaction().context['period'])
+        if context.get('period'):
+            period = Period(context['period'])
             period_for_date = period.find(company_id, date=date,
                 exception=False, test_state=True)
             if period_for_date != period.id:
                 self.date = None
-        elif Transaction().context.get('fiscalyear'):
-            if fiscalyear_for_date != Transaction().context['fiscalyear']:
+        elif context.get('fiscalyear'):
+            if fiscalyear_for_date != context['fiscalyear']:
                 self.date = None
 
     @fields.depends('invoice', 'date', 'batch', 'amount', 'journal')
@@ -686,10 +695,6 @@ class BatchLine(ModelSQL, ModelView):
         pool = Pool()
         Move = pool.get('account.move')
         MoveLine = pool.get('account.move.line')
-        Invoice = pool.get('account.invoice')
-
-        _update_tax_cash_basis = getattr(
-            Invoice, '_update_tax_cash_basis', None)
 
         Move.post([l.move for l in lines
                 if l.move and l.state == 'draft'])
@@ -718,14 +723,10 @@ class BatchLine(ModelSQL, ModelView):
                             reconcile = False
                             break
                 if reconcile:
-                    reconcile_lines, remainder = \
-                        line.invoice.get_reconcile_lines_for_amount(_ZERO)
+                    reconcile_lines, remainder = (
+                        line.invoice.get_reconcile_lines_for_amount(_ZERO))
                     if remainder == _ZERO and move_line in reconcile_lines:
                         MoveLine.reconcile(reconcile_lines)
-            # Compatibility with account_tax_cash
-            if _update_tax_cash_basis:
-                with Transaction().set_context(payment_date=line.date):
-                    Invoice._update_tax_cash_basis([line.invoice])
 
     @classmethod
     def check_unposted_batch_lines_for_account(cls, account_ids=None):
@@ -737,7 +738,8 @@ class BatchLine(ModelSQL, ModelView):
                     ('contra_account', 'in', account_ids),
                     ],
                     ('state', '!=', 'posted'),
-                    ('journal.account_journal.type', '!=', 'situation'),])
+                    ('journal.account_journal.type', '!=', 'situation')
+                ])
         return lines
 
     def set_code(self):
@@ -757,13 +759,7 @@ class BatchLine(ModelSQL, ModelView):
         to_write = []
         for line in lines:
             sequence = Sequence.get('account.batch.line.create')
-            move = line.create_move()
-            move.save()
-            if line.invoice:
-                line._add_payment_line(move)
-                # This check must be run *after* payment lines are added to
-                # have all the payment lines available
-                line._check_invoice_amount()
+            move = line.manage_on_create(line)
             to_write.extend([
                     [line], {'move': move.id, 'sequence': sequence},
                     ])
@@ -772,6 +768,16 @@ class BatchLine(ModelSQL, ModelView):
                 cls.write(*to_write)
         return lines
 
+    def manage_on_create(self, line):
+        move = line.create_move()
+        move.save()
+        if line.invoice:
+            line._add_payment_line(move)
+            # This check must be run *after* payment lines
+            # are added to have all payment lines available
+            line._check_invoice_amount()
+        return move
+
     def create_move(self):
         '''
         Create one move per line and try to reconcile the lines.
@@ -954,7 +960,7 @@ class BatchLine(ModelSQL, ModelView):
             debit=debit,
             credit=credit,
             account=account,
-            party=self.party, # if self.account.party_required else None,
+            party=self.party,  # if self.account.party_required else None,
             second_currency=second_currency,
             amount_second_currency=amount_second_currency,
             maturity_date=self.maturity_date,
@@ -1007,13 +1013,17 @@ class BatchLine(ModelSQL, ModelView):
         for lines, values in zip(actions, actions):
             if cls._check_update_move(values) and not skip_update_move:
                 for line in set(lines):
-                    if line.move:
-                        line._update_move_lines()
-                        if line.invoice:
-                            line._add_payment_line(line.move)
-                            # This check must be run *after* payment lines
-                            # are added to have all payment lines available
-                            line._check_invoice_amount()
+                    line.manage_on_write(line)
+
+    def manage_on_write(self, line):
+        if line.move:
+            line._update_move_lines()
+            if line.invoice:
+                line._add_payment_line(line.move)
+                # This check must be run *after* payment lines
+                # are added to have all payment lines available
+                line._check_invoice_amount()
+        return line
 
     @classmethod
     def _check_update_move(cls, values):
@@ -1153,6 +1163,83 @@ class BatchLine(ModelSQL, ModelView):
         super(BatchLine, cls).delete(lines)
 
 
+class BatchLineTaxCash(metaclass=PoolMeta):
+    __name__ = 'account.batch.line'
+
+    @classmethod
+    def create(cls, vlist):
+        with Transaction().set_context(skip_update_cash_basis=True):
+            return super().create(vlist)
+
+    def manage_on_create(self, line):
+        with Transaction().set_context(payment_date=line.date):
+            return super().manage_on_create(line)
+
+    @classmethod
+    def write(cls, *args):
+        with Transaction().set_context(skip_update_cash_basis=True):
+            super().write(*args)
+
+    def manage_on_write(self, line):
+        with Transaction().set_context(payment_date=line.date):
+            return super().manage_on_write(line)
+
+    def _get_tax_line(self, amount, line_type, tax):
+        '''
+        Mark the tax line as on_cash_basis according to configured periods,
+        i.e. provide a consistent behavior with respect to invoices (that are
+        marked by account_tax_cash from the beginning as on_cash_basis, but
+        without period).
+        '''
+        pool = Pool()
+        Period = pool.get('account.period')
+
+        tax_line = super()._get_tax_line(amount, line_type, tax)
+        company = self.batch.company
+        period_id = Period.find(company.id, date=self.date)
+        period = Period(period_id)
+        if period.is_on_cash_basis(tax):
+            tax_line.on_cash_basis = True
+        return tax_line
+
+    @classmethod
+    def post(cls, lines):
+        pool = Pool()
+        Invoice = pool.get('account.invoice')
+        # Avoid to update cash basis period without the correct context
+        # but do it correctly at the end
+        with Transaction().set_context(skip_update_cash_basis=True):
+            super().post(lines)
+        for line in lines:
+            if line.invoice:
+                with Transaction().set_context(payment_date=line.date):
+                    Invoice._update_tax_cash_basis([line.invoice])
+        cls._update_tax_cash_basis(lines)
+
+    @classmethod
+    def _update_tax_cash_basis(cls, lines):
+        pool = Pool()
+        TaxLine = pool.get('account.tax.line')
+        Date = pool.get('ir.date')
+        Period = pool.get('account.period')
+
+        # Call update_cash_basis grouped per period and ratio only because
+        # group_cash_basis_key already group per move_line.
+        to_update = defaultdict(list)
+        for line in lines:
+            date = line.date or Date.today()
+            if not line.move:
+                continue
+            company = line.batch.company
+            period = Period.find(company.id, date=date)
+            # Fixed ratio, partial payments are handled per line already
+            ratio = 1
+            for line in line.move.lines:
+                to_update[(period, ratio)].extend(line.tax_lines)
+        for (period, ratio), tax_lines in to_update.items():
+            TaxLine.update_cash_basis(tax_lines, ratio, period)
+
+
 class OpenBatchJournalAsk(ModelView):
     'Open Batch Journal Ask'
     __name__ = 'account.batch.open_batch_journal.ask'
@@ -1217,10 +1304,6 @@ class OpenBatchJournal(Wizard):
     open_ = StateAction('account_batch.act_batch_line_form_editable')
 
     def transition_start(self):
-        #if (Transaction().context.get('active_model', '')
-        #        == 'account.journal.period'
-        #        and Transaction().context.get('active_id')):
-        #    return 'open_'
         return 'ask'
 
     def do_open_(self, action):
