@@ -15,14 +15,19 @@ class AccountBatchTestCase(ModuleTestCase):
 
 def suite():
     suite = test_suite()
+    #suite.addTests(doctest.DocFileSuite(
+    #        'scenario_account_batch.rst',
+    #        setUp=doctest_setup, tearDown=doctest_teardown, encoding='utf-8',
+    #        checker=doctest_checker,
+    #        optionflags=doctest.REPORT_ONLY_FIRST_FAILURE))
     suite.addTests(doctest.DocFileSuite(
-            'scenario_account_batch.rst',
+            'scenario_account_batch_tax_cash.rst',
             setUp=doctest_setup, tearDown=doctest_teardown, encoding='utf-8',
-            checker=doctest_checker,
-            optionflags=doctest.REPORT_ONLY_FIRST_FAILURE))
+            checker=doctest_checker,))
+    #optionflags=doctest.REPORT_ONLY_FIRST_FAILURE))
     # Run this test after the scenario, because the scenario needs a fresh
     # database. SQLite doesn't seem to properly rollback the transaction.
     # s. a. https://bugs.tryton.org/issue9133
-    suite.addTests(unittest.TestLoader().loadTestsFromTestCase(
-            AccountBatchTestCase))
+    #suite.addTests(unittest.TestLoader().loadTestsFromTestCase(
+    #        AccountBatchTestCase))
     return suite
